#!/usr/bin/env python3
# Adapted from Eli Bendersky's optimized BF interpreter.

import sys      # for accessing command-line args
import time     # for timing execution in verbose mode

MEMORY_SIZE = 30000

# -----------------------------------------------------
# This function parses a program string,
#      filtering out all non-command characters.
# -----------------------------------------------------

def parse_bf(src_str):
    program = ''
    for c in src_str:
        if c in '><+-.,[]':
            program += c
    return program

# -----------------------------------------------------
# This function precomputes matching brackets that open
#      and close loops, so that the interpreter can
#      jump immediately at run-time.
# -----------------------------------------------------

def compute_jumptable(program):
    pc = 0
    program_size = len(program)
    jumptable = [0] * program_size
    while (pc < program_size):
        instruction = program[pc]
        if instruction == '[':
            bracket_nesting = 1
            seek = pc
            while bracket_nesting > 0 and seek < (program_size - 1):
                seek += 1
                if program[seek] == ']':
                    bracket_nesting -= 1
                elif program[seek] == '[':
                    bracket_nesting += 1
            if bracket_nesting == 0:
                jumptable[pc] = seek
                jumptable[seek] = pc
            else:
                raise ValueError('unmatched [ at pc = ', pc)
        pc += 1
    return jumptable
                
# -----------------------------------------------------
# The interpreter
# -----------------------------------------------------

def interpret(program, jumptable, is_verbose):

    memory   = [0] * MEMORY_SIZE
    pc       = 0
    data_ptr = 0

    while pc < len(program):
        instruction = program[pc]

        if instruction == '>':
            data_ptr += 1
        elif instruction == '<':
            data_ptr -= 1
        elif instruction == '+':
            memory[data_ptr] += 1
        elif instruction == '-':
            memory[data_ptr] -= 1
        elif instruction == '.':
            print(chr(memory[data_ptr]), end='')
        elif instruction == ',':
            memory[data_ptr] = ord(sys.stdin.read(1))
        elif instruction == '[':
            if memory[data_ptr] == 0:
                pc = jumptable[pc]
        elif instruction == ']':
            if memory[data_ptr] != 0:
                pc = jumptable[pc]
        else:
            msg = 'bad character {} at pc = {}'.format(instruction, pc)
            raise ValueError(msg)

        pc += 1

    # Done running the program.  Dump state if verbose.
    if is_verbose:
        print_state(pc, data_ptr, memory)

# -----------------------------------------------------
# The main function takes the command-line args as an
# array.  args[1] is the filename and args[2] is an
# optional '-v' flag to turn on verbose mode.
# -----------------------------------------------------

def main(args):
    bf_file_path = args[1]

    if len(args) > 2 and args[2] == '-v':
        is_verbose = True
    else:
        is_verbose = False

    if is_verbose:
        start_time = time.time()
        print('[>] Parsing: {}'.format(bf_file_path))

    srcfile = open(bf_file_path, 'r')
    program = parse_bf(srcfile.read())
    
    if is_verbose:
        print('[<] Done parsing')
        print('Parsing took: {}s'.format( time.time() - start_time ))
        print()
        print('Length of program: {} chars'.format( len(program) ))
        print('Program: {}'.format( program ))
        print()
        print('[>] Computing jumptable: ')
        start_time = time.time()

    # before running the program, pre-compute the jump targets
    jumptable = compute_jumptable(program)

    if is_verbose:
        print('[<] Done computing jumptable')
        print()
        print(jumptable)
        print()
        print('Jumptable took: {}s'.format( time.time() - start_time ))
        print()
        print('[>] Running {}: '.format(args[0]))
        start_time = time.time()

    interpret(program, jumptable, is_verbose)

    if is_verbose:
        print('[<] Done running')
        print('Execution took: {}s'.format( time.time() - start_time ))
        print()

# --------------------------------------------------------
# These functions print state of machine, for verbose mode
# --------------------------------------------------------

def print_state(pc, data_ptr, memory):
    print()
    print('* pc = ', pc )
    print('* data_ptr = ', data_ptr)
    print_memory(memory)
    print()

def print_memory(memory):
    print('* Nonzero locations in memory:')
    pcount = 0
    for i in range(len(memory)):
        if memory[i] > 0:
            print('[{:5d}] = {:5d}   '.format(i, memory[i]), end='')
            pcount += 1
            if pcount > 0 and pcount % 4 == 0:
                print()

# -----------------------------------------------------
# These lines call the main function with one of three
# test programs, or with the command-line args
# -----------------------------------------------------

if __name__ == "__main__":
    # program_name = 'hello-world.bf'  # 'cell-size.bf' 'factor.bf'
    # is_verbose   = '-v'
    # main(['optimized', program_name, is_verbose])
    main(sys.argv)
